from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Optional

import pandas as pd

from ..dataorigin import DataOrigin
from ..node import OntolocyNode
from ..relationship import OntolocyRelationship


class IngesterBase(ABC):
    node_types: List[OntolocyNode] = []
    rel_types: List[OntolocyRelationship] = []
    relationship_kwargs: dict = {}

    def __init__(
        self,
        data_origin: Optional[DataOrigin] = None,
        private_namespace: Optional[str] = None,
        auto_namespace: Optional[bool] = True,
        assign_data_origin: Optional[bool] = False,
    ) -> None:
        super().__init__()

        self.data_inputs: list = []

        self.node_oriented_dfs: Dict[str, pd.DataFrame] = {}
        self.rel_input_dfs: Dict[str, pd.DataFrame] = defaultdict(dict)
        self.rel_oriented_dfs: Dict[str, pd.DataFrame] = {}

        self.auto_namespace = auto_namespace

        if data_origin is None:
            data_origin_name = self.__class__.__name__

            data_origin = DataOrigin(
                name=data_origin_name,
            )

        if private_namespace:
            self.private_namespace = private_namespace

        else:
            self.private_namespace = None

        if assign_data_origin is True:
            self.data_origin = data_origin

            self.data_origin_dict = {
                "data_origin_name": self.data_origin.name,
                "data_origin_reference": self.data_origin.reference,
                "data_origin_license": self.data_origin.license,
                "data_origin_sharing": self.data_origin.sharing,
            }

        else:
            self.data_origin = None

    def _process_data(self, input_data, private_namespace=None):
        """
        Takes input data in the form of a tuple of dicts:
            Nodes: indexed by node label - with corresponding df for all nodes to merge
            Rels: indexed by rel type - with src, tgt and (optional) rel_props dfs for each

        Drop any duplicates.

        If the ingestor has already been run on different data, append the new data
        """

        self.data_inputs.append(input_data)

        node_oriented_dfs, rel_input_dfs = self._parse(input_data, private_namespace)

        # update node entries
        for label, df in node_oriented_dfs.items():
            # drop any duplicates

            df.drop_duplicates(ignore_index=True, inplace=True)
            self.node_oriented_dfs[label] = pd.concat(
                [self.node_oriented_dfs.get(label, pd.DataFrame()), df]
            )

        # update relationship entries
        for rel_type, entry in rel_input_dfs.items():
            for key, df in entry.items():
                self.rel_input_dfs[rel_type][key] = pd.concat(
                    [
                        self.rel_input_dfs.get(rel_type, {}).get(key, pd.DataFrame()),
                        df,
                    ]
                )

    def _merge_nodes(self) -> None:
        for node_type in self.node_types:
            df = self.node_oriented_dfs[node_type.__primarylabel__]

            if "ontolocy_parser_node_pp" in df.columns:
                nodes = node_type.ingest_df(
                    df.drop(columns=["ontolocy_parser_node_pp"]),
                    self.data_origin,
                )

            else:
                nodes = node_type.ingest_df(
                    df,
                    self.data_origin,
                )

            df["ontolocy_parser_node_pp"] = [
                x.get_primary_property_value() for x in nodes
            ]

    def _generate_relationships(self) -> None:
        for rel_type in self.rel_types:
            dfs = self.rel_input_dfs[rel_type.__relationshiptype__]

            if "props_df" in dfs:
                new_df = dfs.get("props_df").reset_index(drop=True).copy()

            else:
                new_df = pd.DataFrame()

            source_label = rel_type.model_fields["source"].annotation.__primarylabel__

            if "src_df" in dfs and "tgt_df" in dfs:
                # where we only need to specify a single field to match on
                # the src_df should just have one column: 'source'
                # this applies to situations where we aren't creating the nodes in this parser
                # or where the property to match on is deterministic / easy to work out
                if list(dfs["src_df"].columns) == ["source"]:
                    new_df["source"] = (
                        dfs["src_df"]["source"].reset_index(drop=True).copy()
                    )

                # newly created nodes may need to have their ids looked up
                # for example, where the primary property has to be calculated
                elif source_label in self.node_oriented_dfs:
                    src_cols = list(dfs["src_df"].columns) + ["ontolocy_parser_node_pp"]

                    src_df = self.node_oriented_dfs[source_label][src_cols].copy()

                    merged_src_df = (
                        dfs["src_df"]
                        .merge(src_df, how="left")["ontolocy_parser_node_pp"]
                        .copy()
                    )

                    new_df["source"] = merged_src_df

                else:
                    raise (
                        ValueError(
                            "Inappropriately formed source relationship data frames"
                        )
                    )

                target_label = rel_type.model_fields[
                    "target"
                ].annotation.__primarylabel__

                # if we just have a 'target' column, use this to match on
                if list(dfs["tgt_df"].columns) == ["target"]:
                    new_df["target"] = (
                        dfs["tgt_df"]["target"].reset_index(drop=True).copy()
                    )

                # newly created nodes need to have their ids looked up
                elif target_label in self.node_oriented_dfs:
                    tgt_cols = list(dfs["tgt_df"].columns) + ["ontolocy_parser_node_pp"]

                    tgt_df = self.node_oriented_dfs[target_label][tgt_cols].copy()

                    merged_tgt_df = (
                        dfs["tgt_df"]
                        .merge(tgt_df, how="left")["ontolocy_parser_node_pp"]
                        .copy()
                    )

                    new_df["target"] = merged_tgt_df

                else:
                    raise (
                        ValueError(
                            "Inappropriately formed target relationship data frames"
                        )
                    )

            self.rel_oriented_dfs[rel_type.__relationshiptype__] = new_df

    def _merge_relationships(self) -> None:
        for rel_type_class in self.rel_types:

            rel_type = rel_type_class.__relationshiptype__

            df = self.rel_oriented_dfs[rel_type]

            kwargs = self.relationship_kwargs.get(rel_type, {})

            rel_type_class.ingest_df(df, self.data_origin, **kwargs)

    def populate(
        self,
    ) -> None:

        self._merge_nodes()
        self._generate_relationships()
        self._merge_relationships()

    @abstractmethod
    def _parse(self, input_data, private_namespace=None) -> tuple:
        """Returns a tuple where the first entry is a dictionary of node dataframes indexed by label
        The second entry is a dictionary of relationship dataframes indexed by relationship type
            Each relationship entry includes the following:
                * "src_df"
                * "tgt_df"
                * "props_df"
            If the relationship source/target is a node that hasn't just been created,
            then we just expect a single column (source/target).
            Newly created nodes expect sufficient columns to uniquely match the entry in the node df

        """
        raise NotImplementedError



from datetime import datetime
from typing import ClassVar, Optional

from pydantic import HttpUrl

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .intrusionset import IntrusionSet
from .mitreattacksoftware import MitreAttackSoftware
from .mitreattacktechnique import MitreAttackTechnique


class MitreAttackCampaign(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "stix_id"
    __primarylabel__: ClassVar[str] = "MitreAttackCampaign"

    stix_id: str
    stix_type: str
    stix_created: datetime
    stix_modified: datetime
    stix_spec_version: str = "2.1"
    stix_revoked: Optional[bool] = False
    stix_first_seen: datetime
    stix_last_seen: datetime

    attack_spec_version: str
    attack_version: str
    attack_first_seen_citation: str
    attack_last_seen_citation: str
    attack_id: str
    attack_deprecated: Optional[bool] = False

    ref_url: HttpUrl
    name: str
    description: str


#
# OUTGOING RELATIONSHIPS
#


class MitreCampaignUsesTechnique(OntolocyRelationship):
    source: MitreAttackCampaign
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "MITRE_CAMPAIGN_USES_TECHNIQUE"


class MitreCampaignUsesSoftware(OntolocyRelationship):
    source: MitreAttackCampaign
    target: MitreAttackSoftware

    __relationshiptype__: ClassVar[str] = "MITRE_CAMPAIGN_USES_SOFTWARE"


class MitreCampaignAttributedTo(OntolocyRelationship):
    source: MitreAttackCampaign
    target: IntrusionSet

    __relationshiptype__: ClassVar[str] = "MITRE_CAMPAIGN_ATTRIBUTED_TO_INTRUSION_SET"



from abc import abstractmethod
from hashlib import sha256
from pathlib import Path

import requests

from ..ingester import IngesterBase


class OntolocyParser(IngesterBase):
    def detect(self, input_data) -> bool:
        return self._detect(input_data)

    def parse_data(self, input_data, populate=True):
        if self.detect(input_data) is False:
            raise ValueError(
                "Detection suggests input data is not valid for this parser"
            )

        if self.private_namespace is None and self.auto_namespace is True:
            private_namespace = sha256(input_data.encode("utf-8")).hexdigest()

        else:
            private_namespace = self.private_namespace

        self._process_data(input_data, private_namespace)

        if populate is True:
            self.populate()

    def parse_file(self, file_path, populate=True):
        with open(file_path, "r") as f:  # type: ignore [arg-type]
            data = f.read()

        input_data = str(data)

        self.parse_data(input_data, populate=populate)

    def parse_directory(
        self, dir_path, populate: bool = True, path_pattern: str = "**/*"
    ):
        path = Path(dir_path)
        for file_path in path.glob(path_pattern):
            # if we're parsing a directory, don't populate until we've parsed all files
            self.parse_file(file_path, populate=False)

        if populate is True:
            self.populate()

    def parse_url(self, url, populate=True):
        response = requests.get(url)

        data = response.text

        input_data = str(data)

        self.parse_data(input_data, populate=populate)

    @abstractmethod
    def _detect(self, input_data) -> bool:
        """Implement code which returns True if it's valid for this parser
        and returns False otherwise.
        """

        raise NotImplementedError



from click.testing import CliRunner
from ontolocy.cli import cli

import pytest


@pytest.mark.parametrize("prompt_input", ["\n", "no\n"])
@pytest.mark.webtest
def test_mitre_attack_no_populate(prompt_input):

    runner = CliRunner()

    result = runner.invoke(cli, ["parse", "mitre-attack"], input=prompt_input)

    print(result.output)

    assert (
        "Action will create up to 2047 new nodes and 19695 new relationships."
        in result.output
    )

    assert "This might take a few minutes!" in result.output

    assert "Goodbye" in result.output


@pytest.mark.webtest
def test_mitre_attack_bad_prompt():

    runner = CliRunner()

    result = runner.invoke(cli, ["parse", "mitre-attack"], input="bad command\n")

    assert "Error: invalid input" in result.output


@pytest.mark.slow
@pytest.mark.webtest
def test_mitre_attack_populate_url(use_graph):

    runner = CliRunner()

    result = runner.invoke(
        cli,
        [
            "parse",
            "mitre-attack",
            "-u",
            "https://github.com/mitre-attack/attack-stix-data/raw/master/enterprise-attack/enterprise-attack-14.1.json",
        ],
        input="y\n",
    )

    assert "populating the graph..." in result.output
    assert "ingest complete" in result.output

    technique_cypher = """
    MATCH (n:MitreAttackTechnique)
        WHERE  (n.stix_revoked = false OR n.stix_revoked is NULL)
        AND (n.attack_deprecated = false or n.attack_deprecated is NULL)
    RETURN COUNT(DISTINCT n)
    """

    assert use_graph.evaluate_query_single(technique_cypher) == 625



from pathlib import Path

import click

from ontolocy import init_ontolocy
from ontolocy.tools import MitreAttackParser

PARSERS = {
    "mitre-attack": {
        "parser_class": MitreAttackParser,
        "default_url": "https://github.com/mitre-attack/attack-stix-data/raw/master/enterprise-attack/enterprise-attack-14.1.json",
    }
}


@click.group()
def cli():
    pass


@cli.command()
@click.argument("parser_", type=click.Choice(PARSERS.keys()), nargs=1)
@click.option("--filepath", "-p", required=False)
@click.option("--url", "-u", required=False)
@click.option("--force", "-f", type=bool, default=False, required=False)
def parse(parser_, filepath, url, force):
    click.echo(f"parsing {parser_}...")

    parser = PARSERS[parser_]["parser_class"]()

    if filepath:
        path = Path(filepath)

        if path.exists is False:
            click.echo("Doesn't look like the file exists.")
            return

        parser.parse_file(str(path), populate=False)

    elif url:
        parser.parse_url(url, populate=False)

    elif PARSERS[parser_].get("default_url"):
        parser.parse_url(PARSERS[parser_]["default_url"], populate=False)

    else:
        click.echo("Needed a URL or a filepath to data to process.")
        return

    new_nodes = 0
    new_rels = 0

    for node_df in parser.node_oriented_dfs.values():
        new_nodes += len(node_df.index)

    for rel_entry in parser.rel_input_dfs.values():
        if "src_df" in rel_entry:
            new_rels += len(rel_entry["src_df"].index)

    if force is False:

        msg = f"\nAction will create up to {new_nodes} new nodes and {new_rels} new relationships."

        if new_nodes > 1000 or new_rels > 1000:
            msg += "\n\nThis might take a few minutes!"

        msg += "\n\nProceed to populate Neo4j?"

        if click.confirm(msg) is False:
            print("Goodbye")
            return

    print("populating the graph...")
    parser.populate()
    print("ingest complete")


if __name__ == "__main__":
    init_ontolocy()
    cli()



import json

import pandas as pd

from ontolocy import (
    MitreAttackCampaign,
    MitreAttackDataComponent,
    MitreAttackDataComponentDetectsTechnique,
    MitreAttackDataSource,
    MitreAttackDataSourceHasComponent,
    MitreAttackGroup,
    MitreAttackGroupUsesSoftware,
    MitreAttackGroupUsesTechnique,
    MitreAttackMitigation,
    MitreAttackMitigationDefendsAgainstTechnique,
    MitreAttackSoftware,
    MitreAttackTactic,
    MitreAttackTechnique,
    MitreCampaignAttributedTo,
    MitreCampaignUsesSoftware,
    MitreCampaignUsesTechnique,
    MitreSoftwareUsesTechnique,
    MitreSubtechniqueOf,
    MitreTacticIncludesTechnique,
)

from .ontolocy_parser import OntolocyParser


class MitreAttackParser(OntolocyParser):
    node_types = [
        MitreAttackCampaign,
        MitreAttackSoftware,
        MitreAttackTactic,
        MitreAttackTechnique,
        MitreAttackDataSource,
        MitreAttackDataComponent,
        MitreAttackMitigation,
        MitreAttackGroup,
    ]
    rel_types = [
        MitreTacticIncludesTechnique,
        MitreAttackGroupUsesTechnique,
        MitreAttackGroupUsesSoftware,
        MitreSubtechniqueOf,
        MitreCampaignUsesSoftware,
        MitreCampaignAttributedTo,
        MitreCampaignUsesTechnique,
        MitreSoftwareUsesTechnique,
        MitreAttackMitigationDefendsAgainstTechnique,
        MitreAttackDataComponentDetectsTechnique,
        MitreAttackDataSourceHasComponent,
    ]

    def _detect(self, input_data) -> bool:
        try:
            stix_json = json.loads(input_data)

        except json.decoder.JSONDecodeError:
            print("json decode error")
            return False

        if "x_mitre_attack_spec_version" in stix_json["objects"][0]:
            return True

        else:
            return False

    def _stix_objects_to_df(self, stix_data, stix_types):
        """Pull out stix objects of a certain type"""

        all_allowed_columns = set(
            [
                "stix_id",
                "stix_type",
                "stix_spec_version",
                "stix_created",
                "stix_modified",
                "stix_revoked",
                "attack_version",
                "attack_spec_version",
                "attack_id",
                "ref_url",
                "name",
                "description",
                "attack_subtechnique",
                "attack_shortname",
                "stix_first_seen",
                "stix_last_seen",
                "attack_first_seen_citation",
                "attack_last_seen_citation",
                "kill_chain_phase_list",
                "aliases",
                "x_mitre_data_source_ref",
                "attack_deprecated",
            ]
        )

        attack_objs = [
            x
            for x in stix_data["objects"]
            if x["type"] in stix_types
            # and not x.get("x_mitre_deprecated")
            # and not x.get("revoked")
        ]

        # pull out associated MITRE ATT&CK tactics the MITRE ATT&CK reference (web address)
        for attack_obj in attack_objs:

            if attack_obj.get("kill_chain_phases"):
                attack_obj["kill_chain_phase_list"] = []

                for phase in attack_obj.get("kill_chain_phases", []):
                    if phase["kill_chain_name"] == "mitre-attack":
                        attack_obj["kill_chain_phase_list"].append(phase["phase_name"])

            for ref in attack_obj.get("external_references", []):
                if ref["source_name"] == "mitre-attack":
                    attack_obj["attack_id"] = ref["external_id"]
                    attack_obj["ref_url"] = ref["url"]
                    break

        df = pd.DataFrame.from_records(attack_objs).rename(
            columns={
                "id": "stix_id",
                "type": "stix_type",
                "created": "stix_created",
                "modified": "stix_modified",
                "spec_version": "stix_spec_version",
                "revoked": "stix_revoked",
                "x_mitre_version": "attack_version",
                "x_mitre_is_subtechnique": "attack_subtechnique",
                "x_mitre_attack_spec_version": "attack_spec_version",
                "x_mitre_shortname": "attack_shortname",
                "first_seen": "stix_first_seen",
                "last_seen": "stix_last_seen",
                "x_mitre_first_seen_citation": "attack_first_seen_citation",
                "x_mitre_last_seen_citation": "attack_last_seen_citation",
                "x_mitre_deprecated": "attack_deprecated",
            }
        )

        all_columns = set(df.columns)

        drop_columns = all_columns.difference(all_allowed_columns)

        df_subset = df.drop(columns=list(drop_columns))

        return df_subset

    def _stix_rels_to_df(self, stix_data):
        rel_objs = [x for x in stix_data["objects"] if x["type"] == "relationship"]

        rel_df = pd.DataFrame.from_records(rel_objs)

        return rel_df

    def _parse(self, input_data, private_namespace=None) -> tuple:
        stix_json = json.loads(input_data)

        #
        # Nodes
        #

        node_dfs = {}

        tactics_df = self._stix_objects_to_df(stix_json, ["x-mitre-tactic"])

        tactics_dict = (
            tactics_df[["attack_shortname", "stix_id"]]
            .set_index("attack_shortname")["stix_id"]
            .to_dict()
        )

        node_dfs[MitreAttackTactic.__primarylabel__] = tactics_df.copy()

        techniques_df = self._stix_objects_to_df(stix_json, ["attack-pattern"])

        node_dfs[MitreAttackTechnique.__primarylabel__] = techniques_df.drop(
            columns=["kill_chain_phase_list"]
        ).copy()

        groups_df = self._stix_objects_to_df(stix_json, ["intrusion-set"])

        node_dfs[MitreAttackGroup.__primarylabel__] = groups_df.drop(
            columns=["aliases"]
        ).copy()

        campaigns_df = self._stix_objects_to_df(stix_json, ["campaign"])

        node_dfs[MitreAttackCampaign.__primarylabel__] = campaigns_df.drop(
            columns=["aliases"]
        ).copy()

        software_df = self._stix_objects_to_df(stix_json, ["tool", "malware"])

        node_dfs[MitreAttackSoftware.__primarylabel__] = software_df.copy()

        mitigations_df = self._stix_objects_to_df(stix_json, ["course-of-action"])

        node_dfs[MitreAttackMitigation.__primarylabel__] = mitigations_df.copy()

        datasources_df = self._stix_objects_to_df(stix_json, ["x-mitre-data-source"])

        node_dfs[MitreAttackDataSource.__primarylabel__] = datasources_df.copy()

        datacomponents_df = self._stix_objects_to_df(
            stix_json, ["x-mitre-data-component"]
        )

        if datacomponents_df.empty is False:

            node_dfs[
                MitreAttackDataComponent.__primarylabel__
            ] = datacomponents_df.drop(columns=["x_mitre_data_source_ref"]).copy()

        else:
            node_dfs[MitreAttackDataComponent.__primarylabel__] = pd.DataFrame()

        #
        # Relationships
        #

        rel_dfs = {}

        all_relationships_df = self._stix_rels_to_df(stix_json).rename(
            columns={"source_ref": "source", "target_ref": "target"}
        )

        # Techniques to Tactics

        tactic_technique_df = pd.DataFrame()

        for technique_row in techniques_df.iterrows():
            target = technique_row[1].stix_id

            sources = [tactics_dict[x] for x in technique_row[1].kill_chain_phase_list]

            entry_df = pd.DataFrame()
            entry_df["source"] = sources
            entry_df["target"] = target

            tactic_technique_df = pd.concat([tactic_technique_df, entry_df])

        rel_dfs[MitreTacticIncludesTechnique.__relationshiptype__] = {
            "src_df": tactic_technique_df[["source"]].copy(),
            "tgt_df": tactic_technique_df[["target"]].copy(),
        }

        # Techniques to Subtechniques

        techniques_to_subtechniques_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("attack-pattern")
            & all_relationships_df.target.str.startswith("attack-pattern")
            & all_relationships_df.relationship_type.str.match("subtechnique-of")
        ]

        rel_dfs[MitreSubtechniqueOf.__relationshiptype__] = {
            "src_df": techniques_to_subtechniques_df[["source"]].copy(),
            "tgt_df": techniques_to_subtechniques_df[["target"]].copy(),
        }

        # Group to Technique

        group_to_technique_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("intrusion-set")
            & all_relationships_df.target.str.startswith("attack-pattern")
        ]

        rel_dfs[MitreAttackGroupUsesTechnique.__relationshiptype__] = {
            "src_df": group_to_technique_df[["source"]].copy(),
            "tgt_df": group_to_technique_df[["target"]].copy(),
        }

        # Group to Software

        group_to_software_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("intrusion-set")
            & (
                all_relationships_df.target.str.startswith("tool")
                | all_relationships_df.target.str.startswith("malware")
            )
        ]

        rel_dfs[MitreAttackGroupUsesSoftware.__relationshiptype__] = {
            "src_df": group_to_software_df[["source"]].copy(),
            "tgt_df": group_to_software_df[["target"]].copy(),
        }

        # Campaigns to Software

        campaign_to_software_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("campaign")
            & (
                all_relationships_df.target.str.startswith("tool")
                | all_relationships_df.target.str.startswith("malware")
            )
        ]

        rel_dfs[MitreCampaignUsesSoftware.__relationshiptype__] = {
            "src_df": campaign_to_software_df[["source"]].copy(),
            "tgt_df": campaign_to_software_df[["target"]].copy(),
        }

        # Campaigns to Techniques

        campaign_to_technique_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("campaign")
            & all_relationships_df.target.str.startswith("attack-pattern")
        ]

        rel_dfs[MitreCampaignUsesTechnique.__relationshiptype__] = {
            "src_df": campaign_to_technique_df[["source"]].copy(),
            "tgt_df": campaign_to_technique_df[["target"]].copy(),
        }

        # Campaigns to Intrusion Sets

        campaign_to_intrusion_set_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("campaign")
            & all_relationships_df.target.str.startswith("intrusion-set")
            & all_relationships_df.relationship_type.str.match("attributed-to")
        ]

        rel_dfs[MitreCampaignAttributedTo.__relationshiptype__] = {
            "src_df": campaign_to_intrusion_set_df[["source"]].copy(),
            "tgt_df": campaign_to_intrusion_set_df[["target"]].copy(),
        }

        # Software to Techniques

        software_to_technique_df = all_relationships_df.loc[
            (
                all_relationships_df.source.str.startswith("tool")
                | all_relationships_df.source.str.startswith("malware")
            )
            & all_relationships_df.target.str.startswith("attack-pattern")
        ]

        rel_dfs[MitreSoftwareUsesTechnique.__relationshiptype__] = {
            "src_df": software_to_technique_df[["source"]].copy(),
            "tgt_df": software_to_technique_df[["target"]].copy(),
        }

        # Mitigation to Techniques

        mitigation_to_technique_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("course-of-action")
            & all_relationships_df.target.str.startswith("attack-pattern")
            & all_relationships_df.relationship_type.str.match("mitigates")
        ]

        rel_dfs[MitreAttackMitigationDefendsAgainstTechnique.__relationshiptype__] = {
            "src_df": mitigation_to_technique_df[["source"]].copy(),
            "tgt_df": mitigation_to_technique_df[["target"]].copy(),
        }

        # Data Component to Techniques

        mitigation_to_technique_df = all_relationships_df.loc[
            all_relationships_df.source.str.startswith("x-mitre-data-component")
            & all_relationships_df.target.str.startswith("attack-pattern")
            & all_relationships_df.relationship_type.str.match("detects")
        ]

        rel_dfs[MitreAttackDataComponentDetectsTechnique.__relationshiptype__] = {
            "src_df": mitigation_to_technique_df[["source"]].copy(),
            "tgt_df": mitigation_to_technique_df[["target"]].copy(),
        }

        # Data Source to Data Component

        if datacomponents_df.empty is False:

            data_source_to_component_df = pd.DataFrame()

            data_source_to_component_df["source"] = datacomponents_df[
                "x_mitre_data_source_ref"
            ]
            data_source_to_component_df["target"] = datacomponents_df["stix_id"]

            rel_dfs[MitreAttackDataSourceHasComponent.__relationshiptype__] = {
                "src_df": data_source_to_component_df[["source"]].copy(),
                "tgt_df": data_source_to_component_df[["target"]].copy(),
            }

        return node_dfs, rel_dfs
